#
# (c) 2013 Mandriva, http://www.mandriva.com
#
# This file is part of Mandriva Management Console (MMC).
#
# MMC is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# MMC is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with MMC.  If not, see <http://www.gnu.org/licenses/>.
#
import urllib

"""
MDS davical plugin for the MMC agent.
"""
import os
import logging
import urlparse

from mmc.core.version import scmRevision
from mmc.plugins.davical.config import DavicalPluginConfig

from caldavclientlibrary.client.account import CalDAVAccount
from caldavclientlibrary.protocol.url import URL
from caldavclientlibrary.protocol.webdav.definitions import davxml
from caldavclientlibrary.protocol.webdav.ace import ACE


VERSION = "0.6.2"
APIVERSION = "0:0:0"
REVISION = scmRevision("$Rev$")

def getVersion(): return VERSION
def getApiVersion(): return APIVERSION
def getRevision(): return REVISION

def activate():
    config = DavicalPluginConfig('davical')
    logger = logging.getLogger()
    if config.disabled:
        msg = "disabled by configuration"
        logger.warning("Plugin davical: " + msg + ".")
        return False

    if Davical().account.getPrincipal() is None:
        logger.warn('No user: "%s" or wrong password, check your configuration', config.user)
        return False

    logger.debug("server: %s " % config.server)
    logger.debug("user: %s " % config.user)
    logger.debug("pswd: %s " % config.pswd)
    return True

def getCalendars(user = None, pswd = None):
    return Davical(user, pswd).getCalendars()

def getCalendarsForUser(user):
    logging.debug("getCalendarsForUser: %s", user)
    return Davical().getCalendarsForUser(user)

def delCalendar(calendar, user = None, pswd = None):
    logging.debug("delCalendar: %s" % calendar)
    return Davical(user, pswd).delCalendars(calendar)

def addCalendar(name, desc="", user = None, pswd = None):
    logging.debug("addCalendar: %s" % name)
    return Davical(user, pswd).addCalendar(name, desc)

def getACL(calendar, user = None, pswd = None):
    logging.debug("getACL: %s" % calendar)
    return Davical(user, pswd).getACL(calendar)

def setACL(calendar, acls, user = None, pswd = None):
    return Davical(user, pswd).setACL(calendar, acls)

class Davical:
    def __init__(self, user = None, pswd = None):
        self.config = DavicalPluginConfig('davical')

        self.logger = logging.getLogger()

        ssl = self.config.server.startswith("https://")
        splits = urlparse.urlsplit(self.config.server)
        server = splits.scheme + "://" + splits.netloc
        path = splits.path
        server = server[8:] if ssl else server[7:]

        if user is None:
            user = self.config.user
        if pswd is None:
            pswd = self.config.pswd

        self.logger.debug("server: %s", server)
        self.logger.debug("path: %s", path)
        self.logger.debug("user: %s", self.config.user)
        self.logger.debug("ssl: %s", ssl)


        self.account = CalDAVAccount(server,
                            #port=port,
                            ssl=ssl,
                            user=user,
                            pswd=pswd,
                            root=path,
                            principal=None,
                            logging=False)
                            #logging=True) #uncomment to display http request
        self.allowListCalendars()

    def allowListCalendars(self, principal=None):
        self.logger.debug("allowListCalendars")

        homepath = self.account.getPrincipal().homeset[0].path
        results, bad = self.account.session.getProperties(URL(url=homepath),
                                                          (davxml.acl, ))
        currentAces = ACE.parseFromACL(results[davxml.acl])
        currentAces = [ace for ace in currentAces if ace.canChange()]

        for ace in currentAces:
            self.logger.debug("acl loop, ace.data: %s", ace.data)
            # update defaults rights
            if ace.getPrincipal() == davxml.authenticated:
                self.logger.debug("default rule matched")
                if not davxml.read in ace.privs:
                    ace.grant =  True
                    ace.privs += (davxml.read,)
                    break

        self.account.session.setACL(URL(url=homepath), currentAces)



    def getCalendars(self, principal=None):
        if principal is None:
            #try the default
            principal = self.account.getPrincipal()
            if principal is None:
                #should not appen
                return None

        self.logger.debug("getCalendars> principal: %s", principal)

        principal.loadDetails(True)
        calendars = principal.listCalendars()
        cals = []
        for calendar in calendars:
            self.logger.debug(calendar)
            self.logger.debug(calendar.path)
            name = calendar.getDisplayName()
            if name:
                self.logger.debug("  Display Name: %s" % (name,))
            desc = calendar.getDescription()
            if desc:
                self.logger.debug("  Description: %s" % (desc,))
            cals.append((calendar.path, name, desc))
        return cals

    def getCalendarsForUser(self, user):
        homepath = self.account.getPrincipal().homeset[0].path
        self.logger.debug("getCalendarsForUser> path: %s", homepath)
        path2 = os.path.dirname(homepath)
        path2 = os.path.dirname(path2)
        ppath = os.path.join(path2, user)
        if not ppath.endswith("/"):
            ppath += "/"
        self.logger.debug("getCalendarsForUser> URL(url=ppath): %s", URL(url=ppath))

        principal = self.account.getPrincipal(URL(url=ppath))
        self.logger.debug("getCalendarsForUser> principal: %s", principal)
        try:
            return self.getCalendars(principal)
        except IndexError:
            return []

    def addCalendar(self, calName, calDescription=""):
        self.logger.debug("add calendar name: %s description: %s", calName, calDescription)
        homepath = self.account.getPrincipal().homeset[0].path
        if not calName.startswith("/"):
            calName = os.path.join(homepath, calName)
        if not calName.endswith("/"):
            calName += "/"
        newcal = URL(url=calName)
        self.account.session.makeCalendar(newcal,
                                          displayname=calName,
                                          description=calDescription)


    def delCalendars(self, calendar):
        self.logger.debug("delete calendar: %s" % calendar)
        newcal = URL(url=calendar)
        self.account.session.deleteResource(newcal)


    def stripPrincipalName(self, principal_name):
        principal_name = urllib.unquote(principal_name)
        #pop trailing '/'
        principal_name = os.path.split(principal_name)[0]
        #get principal name i.e. drop leading path
        principal_name = os.path.split(principal_name)[1]
        self.logger.debug("principal_name %s", principal_name)
        return principal_name

    def getACL(self, calendar):
        self.logger.debug("getACL for calendar: %s" % calendar)

        results, bad = self.account.session.getProperties(URL(url=calendar),
                                                          (davxml.acl, ))
        aces = ACE.parseFromACL(results[davxml.acl])
        aces = [ace for ace in aces if ace.canChange()]
        self.logger.debug("rootPath %s", self.account.session.rootPath)
        access = []
        accessDict = {}
        for ace in aces:
            self.logger.debug("principal_type %s", ace.getPrincipal())
            self.logger.debug("data %s", ace.data)
            self.logger.debug("privs %s", ace.privs)
            principal_type = ace.getPrincipal()
            principal_name = str(ace.data) if ace.data else ""

            right = ""
            if davxml.write in ace.privs:
                right = "rw"
            elif davxml.read in ace.privs:
                right = "ro"

            if principal_type == davxml.href:
                principal_name = self.stripPrincipalName(principal_name)
            if principal_type == davxml.authenticated:
                principal_name = "default"
            if principal_type in (davxml.href, davxml.authenticated):
                access.append((principal_type,
                              principal_name,
                              right))
                accessDict[principal_name] = right

        #return access
        return accessDict

    def setPrivs(self, priv):
        if priv == "":
            return True, ()
        elif priv == "ro":
            return True, (davxml.read,)
        elif priv == "rw":
            return True, (davxml.read, davxml.write)
        else:
            return False, ()

    def setACL(self, calendar, acls):
        self.logger.debug("setACL for calendar: %s", calendar)
        for acl in acls:
            self.logger.debug("    ACL: %s, %s", acl, acls[acl])

        results, bad = self.account.session.getProperties(URL(url=calendar),
                                                          (davxml.acl, ))
        currentAces = ACE.parseFromACL(results[davxml.acl])
        currentAces = [ace for ace in currentAces if ace.canChange()]
        newAces = []
        rootpath = self.account.session.rootPath

        for ace in currentAces:
            self.logger.debug("acl loop, ace.data: %s", ace.data)
            principal_name = str(ace.data) if ace.data else ""

            # don't do anything
            if principal_name == davxml.owner:
                self.logger.debug("skip owner")
                continue

            # update defaults for authenticated users
            if ace.getPrincipal() == davxml.authenticated and "default" in acls:
                self.logger.debug("default rule matched")
                ace.grant, ace.privs = self.setPrivs(acls["default"])
                newAces.append(ace)
                del acls["default"]

            principal_name = self.stripPrincipalName(principal_name)
            # some unmanaged rules, probably added using the davical web interface
            if principal_name not in acls:
                self.logger.info("skip unmanaged principal %s with name: '%s'",
                                 ace.getPrincipal(),
                                 principal_name)
                newAces.append(ace)
                continue
            # update any existing rules
            else:
                priv = acls[principal_name]
                ace.grant, ace.privs = self.setPrivs(priv)
                if ace.grant:
                    self.logger.debug("update rights to %s for user %s", priv, principal_name)
                    newAces.append(ace)
                else:
                    self.logger.debug("revoke rights to user %s", principal_name)
                del acls[principal_name]

        # check if some new rules have to be added
        if len(acls) > 0:
            self.logger.debug("found %s new rule(s)", len(acls))
            #loop through all remaning acls (i.e. new users)
            for principal, acl in acls.items():
                if acl != "revoke":
                    self.logger.debug("pushing new acl for: %s", principal)
                    newAce = ACE()
                    newPrincipal = principal

                    if not newPrincipal.startswith("/"):
                        newPrincipal = os.path.join(rootpath.toString(), newPrincipal)
                    if not newPrincipal.endswith("/"):
                        newPrincipal += "/"

                    newAce.setPrincipal(davxml.href,
                                        newPrincipal)
                    newAce.grant, newAce.privs = self.setPrivs(acl)
                    self.logger.debug("new principal: %s named: '%s' with right: %s",
                                      newAce.getPrincipal(),
                                      newAce.data,
                                      newAce.privs)
                    newAces.append(newAce)

        self.account.session.setACL(URL(url=calendar), newAces)
